"""Git operations"""
from __future__ import annotations
from dataclasses import dataclass
from typing import List, Mapping, Tuple

from parse import compile, Parser, Result

from publicator import shell
from semmy import Semver


class RemoteParser:
    parser: Parser

    def __init__(self) -> None:
        self.parser = compile("git@{server}:{owner}/{name}.git")

    def parse(self, remote: str) -> Mapping[str, str]:
        result = self.parser.parse(remote)

        if not isinstance(result, Result):
            return {}

        named: Mapping[str, str] = result.named
        return named


@dataclass(frozen=True)  # pragma: no mutate
class Repo:
    server: str = ""
    owner: str = ""
    name: str = ""

    @classmethod
    def from_remote(cls) -> Repo:
        parser = RemoteParser()
        remote = shell.run("git remote get-url --push origin").pop()
        values = parser.parse(remote)

        return cls(server=values.get("server", ""), owner=values.get("owner", ""), name=values.get("name", ""))

    @property
    def is_github(self) -> bool:
        return "github.com" in self.server


def current_branch() -> str:
    return shell.run("git symbolic-ref --short HEAD").pop()


def release_branches() -> Tuple[str, str]:
    return ("main", "master")


def status() -> List[str]:
    return shell.run("git status --porcelain")


def is_working_directory_clean() -> bool:
    return len(status()) == 0


def stash() -> List[str]:
    return shell.run("git stash -u")


def pull() -> List[str]:
    return shell.run("git pull --rebase")


def pop() -> List[str]:
    return shell.run("git stash pop")


def add() -> List[str]:
    return shell.run("git add pyproject.toml")


def commit(message: str) -> List[str]:
    return shell.run(f'git commit -m "{message.strip()}"')


def create_tag(version: Semver, message: str) -> List[str]:
    return shell.run(f'git tag -a {version} -m "{message.strip()}"')


def push() -> List[str]:
    return shell.run("git push --follow-tags")



from urllib.parse import urlencode
from publicator.git import Repo
from semmy import Semver


class ReleaseException(Exception):
    pass


def new_release_url(repo: Repo, tag: Semver, title: str = "", body: str = "") -> str:
    if not repo.is_github:
        raise ReleaseException("Current repository is not hosted on github.com")

    base_url = f"https://{repo.server}/{repo.owner}/{repo.name}/releases/new?"
    query = {"tag": tag, "title": title, "body": body, "prerelease": int(tag.is_pre_release())}

    return base_url + urlencode(query)



import os
from typing import Optional

import typer

import publicator
from publicator import config, git, github, poetry
from semmy import Semver

preview = os.environ.get("PUBLICATOR_PREVIEW")
configuration = config.factory()
app = typer.Typer(name=publicator.name)


def version_callback(value: bool) -> None:
    if not value:
        return

    typer.secho(f"🦄 {publicator.version()}", fg=typer.colors.BRIGHT_BLUE, bold=True)
    raise typer.Exit()


@app.command()
def cli(
    bump: str = typer.Argument(
        ...,
        metavar="version",
        help="can be a valid semver or one of: patch, minor, major, prepatch, preminor, premajor, prerelease",
    ),
    version: Optional[bool] = typer.Option(None, "--version", "-V", callback=version_callback, is_eager=True),
    repository: Optional[str] = typer.Option(
        default=configuration.get("repository"),
        metavar="name",
        help="Custom repository for publishing (must be specified in pyproject.toml)",
    ),
    any_branch: bool = typer.Option(
        default=configuration.get("any-branch", False), help="Allow publishing from any branch"
    ),
    clean: bool = typer.Option(
        default=configuration.get("clean", True), help="Ensure you're working with the latest changes"
    ),
    tag: bool = typer.Option(default=configuration.get("tag", True), help="Create a new tag for Git"),
    publish: bool = typer.Option(
        default=configuration.get("publish", True), help="Publish the package to the registry"
    ),
    push: bool = typer.Option(default=configuration.get("push", True), help="Push commits and tags to Git"),
    test_script: str = typer.Option(
        default=configuration.get("test-script", "pytest -x --assert=plain"),
        help="Name of the test script to run under the current virtual environment",
    ),
    template: str = typer.Option(
        default=configuration.get("template", "release: %s"),
        help="Commit message template (`%s` will be replaced with the new version tag)",
    ),
    release_draft: bool = typer.Option(
        default=configuration.get("release-draft", True),
        help="Opens a pre-filled GitHub release page with browser if the current project is hosted on GitHub",
    ),
) -> None:
    """
    Handles publishing a new Python package via Poetry safely and conveniently.
    """

    if not any_branch:
        verify_branch()

    if clean:
        clean_up()
        install_dependencies()

    if test_script:
        run_tests(test_script)

    new_version = bump_version(bump)
    commit_changes(new_version, template)

    if tag:
        create_tag(new_version)

    if publish:
        build_package()
        publish_package(repository)

    if push:
        push_changes()

    if release_draft:
        draft_new_release(new_version)

    success("Published the new package version. Cheers!")


def info(message: str) -> None:
    typer.secho(f"📢 {message}", fg=typer.colors.BRIGHT_BLUE)


def success(message: str) -> None:
    typer.secho(f"🎉 {message}", fg=typer.colors.BRIGHT_GREEN, bold=True)


def warn(message: str) -> None:
    typer.secho(f"⚠️ {message}", fg=typer.colors.BRIGHT_YELLOW)


def fatal(message: str, exit_code: int = 1) -> None:
    typer.secho(f"❌ {message}", fg=typer.colors.BRIGHT_RED, bold=True)
    raise typer.Exit(code=exit_code)


def draft_new_release(tag: Semver) -> None:
    repo = git.Repo.from_remote()

    info("Opening GitHub release draft in browser")

    if preview:
        return

    try:
        url = github.new_release_url(repo, tag, title=f"Version {tag}", body="Write here")
        typer.launch(url)
    except github.ReleaseException as error:
        warn(f"Skipping release draft. Reason: {error}")


def push_changes() -> None:
    info("Pushing changes to Git")
    if not preview:
        git.push()


def publish_package(repository: Optional[str]) -> None:
    info("Publishing the package to repository")
    if not preview:
        poetry.publish(repository)


def build_package() -> None:
    info("Building the package")
    if not preview:
        poetry.build()


def create_tag(version: Semver) -> None:
    info(f"Creating a new tag {version} from HEAD")
    if not preview:
        git.create_tag(version, message=f"Version {version}")


def commit_changes(semver: Semver, template: str) -> None:
    message = template % (semver)
    info(f'Committing changes with message: "{message}"')

    if not preview:
        poetry.ok()
        git.add()
        git.commit(message)


def bump_version(version: str) -> Semver:
    current_version = poetry.version()

    if preview:
        return current_version

    next_version = poetry.bump(version)
    info(f"Bumped version from {current_version} to {next_version}")

    return next_version


def run_tests(script: str) -> None:
    info(f"Running tests with '{script}'")
    if not preview:
        poetry.run(script)


def install_dependencies() -> None:
    info("Reinstalling dependencies")
    if not preview:
        poetry.install()


def clean_up() -> None:
    if git.is_working_directory_clean():
        return

    if not preview:
        info("Resetting working directory to a clean state")
        git.stash()
        git.pull()
        git.pop()


def verify_branch() -> None:
    current_branch = git.current_branch()
    release_branches = git.release_branches()

    if current_branch not in release_branches:
        fatal(f"Current checked out branch {current_branch} is not a release branch {release_branches}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['publicator']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'parse>=1.19.0,<2.0.0',
 'semmy>=0.2.0,<0.3.0',
 'tomli==2.0.1',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['publicator = publicator.cli:app']}

setup_kwargs = {
    'name': 'publicator',
    'version': '1.1.1',
    'description': 'A better poetry publish experience.',
    'long_description': '<h1>🗞 Publicator</h1>\n\n> A better `poetry publish` experience.\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/publicator)\n![PyPI](https://img.shields.io/pypi/v/publicator)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/publicator)\n![PyPI - License](https://img.shields.io/pypi/l/publicator)\n[![Application Test Suite](https://github.com/nikoheikkila/publicator/actions/workflows/python-package.yml/badge.svg)](https://github.com/nikoheikkila/publicator/actions/workflows/python-package.yml)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/publicator)\n\nWhile [Poetry](https://python-poetry.org) finally brings us a sane solution for publishing and maintaining Python packages, many developers crave for a more _enhanced_ and safer user experience. Publicator aims to offer a convenient method for publishing your everyday libraries.\n\nPublicator has been inspired by Sindre Sorhus\' excellent [`np`](https://github.com/sindresorhus/np) package for Node.js ecosystem and graciously funded by [**Futurice Spice Program**](https://spiceprogram.org).\n\n<h2>Table of Contents</h2>\n\n* [Features](#features)\n* [Prerequisites](#prerequisites)\n* [Install](#install)\n* [Usage](#usage)\n  * [Configuration](#configuration)\n  * [Preview Mode (Dry-Run)](#preview-mode-dry-run)\n  * [Shell Completion](#shell-completion)\n* [Contributing](#contributing)\n\n## Features\n\n* Ensures you are publishing from your release branch (`main` and `master` by default)\n* Ensures the working directory is clean and latest changes are pulled\n* Reinstalls dependencies to ensure your project works with the latest dependency tree\n* Ensures your Python version is supported by the project and its dependencies\n* Runs the tests with custom test script\n* Bumps the version in `pyproject.toml` and creates a Git tag based on it\n* Publishes the new version to [Python Package Index](https://pypi.org) or custom repository\n* Pushes commits and tags (newly & previously created) to your Git server\n* If the project is hosted on GitHub opens a prefilled GitHub Releases draft after publishing\n* Fully configurable via command-line arguments or the `pyproject.toml` file\n* See what will be executed with preview mode, without pushing or publishing anything remotely\n\n## Prerequisites\n\n* **Python 3.8** or later\n* **Poetry 1.1** or later\n* **Git 2.11** or later\n\n## Install\n\nInstall or run directly using `pipx`, which manages an isolated virtual environment for you.\n\n```sh\npipx install publicator\npipx run publicator <version>\n```\n\nAlternatively, add it as dependency to your Poetry project.\n\n```sh\npoetry add --dev publicator\npoetry run publicator <version>\n```\n\n## Usage\n\nPublicator takes one command-line argument indicating the suitable version bump. It follows semantic versioning rules accurately.\n\n```sh\n# Release a new patch version (e.g. 1.0.0 -> 1.0.1)\npublicator patch\n\n# Release a new minor version (e.g. 1.0.1 -> 1.1.0)\npublicator minor\n\n# Release a new major version (e.g. 1.1.0 -> 2.0.0)\npublicator major\n```\n\nRun `publicator --help` to see the full list of supported options:\n\n```plain\n$ publicator --help\n\nUsage: publicator [OPTIONS] version\n\n  Handles publishing a new Python package via Poetry safely and conveniently.\n\nArguments:\n  version  can be a valid semver or one of: patch, minor, major, prepatch,\n           preminor, premajor, prerelease  [required]\n\nOptions:\n  -V, --version\n  --repository name               Custom repository for publishing (must be\n                                  specified in pyproject.toml)\n  --any-branch / --no-any-branch  Allow publishing from any branch  [default:\n                                  no-any-branch]\n  --clean / --no-clean            Ensure you\'re working with the latest\n                                  changes  [default: clean]\n  --tag / --no-tag                Create a new tag for Git  [default: tag]\n  --publish / --no-publish        Publish the package to the registry\n                                  [default: publish]\n  --push / --no-push              Push commits and tags to Git  [default:\n                                  push]\n  --test-script TEXT              Name of the test script to run under the\n                                  current virtual environment  [default:\n                                  pytest -x --assert=plain]\n  --template TEXT                 Commit message template (`%s` will be\n                                  replaced with the new version tag)\n                                  [default: release: %s]\n  --release-draft / --no-release-draft\n                                  Opens a pre-filled GitHub release page with\n                                  browser if the current project is hosted on\n                                  GitHub  [default: release-draft]\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n  --help                          Show this message and exit.\n```\n\n### Configuration\n\nPublicator follows the pleasantly Pythonic way of specifying the configuration within `pyproject.toml` file. Below are the default configuration values.\n\n```toml\n[tool.publicator]\nany-branch    = false\nclean         = true\npublish       = true\npush          = true\nrelease-draft = true\ntag           = true\ntemplate      = "release: %s"\n```\n\nValues passed as command-line arguments take precedence over configuration file values.\n\nConfiguration enables for more granular usage. For example, in CI/CD pipelines you might want to disable publishing the package to registry or disable creating Git tags depending on your use case.\n\n### Preview Mode (Dry-Run)\n\nIf you\'d rather skip on everything and check what would be executed, you can activate a preview mode via environment variable like so:\n\n```sh\nPUBLICATOR_PREVIEW=true publicator <version>\n```\n\n### Shell Completion\n\nPublicator stands on the shoulders of [Typer](https://typer.tiangolo.com), which is a robust CLI library for Python. You can generate <kbd>TAB</kbd> completions for common shells such as Bash, ZSH, Fish, and Powershell.\n\n```sh\npublicator --install-completion <shell>\n```\n\n## Contributing\n\nSee [**here**](CONTRIBUTING.md) for instructions.\n',
    'author': 'Niko Heikkilä',
    'author_email': 'niko.heikkila@futurice.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/publicator/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
